@@ -131,6 +131,7 @@ enum
   REQUEST_RESOURCE,
   FREE_RESOURCE,
   SRC_PAD_ADDED,
+  RESTART_SOURCE,
   SIGNAL_COUNT
 };
 
@@ -719,6 +720,24 @@ tf_stream_class_init (TfStreamClass *klass)
           NULL, NULL,
           _tf_marshal_VOID__OBJECT_BOXED,
           G_TYPE_NONE, 2, GST_TYPE_PAD, FS_TYPE_CODEC);
+
+
+  /**
+   * TfStream::restart-source:
+   * @stream: the stream
+   *
+   * This is emitted when there is a caps change and the source should be
+   * restarted to take this into account.
+   */
+
+  signals[RESTART_SOURCE] =
+      g_signal_new ("restart-source",
+          G_OBJECT_CLASS_TYPE (klass),
+          G_SIGNAL_RUN_LAST,
+          0,
+          NULL, NULL,
+          _tf_marshal_VOID__VOID,
+          G_TYPE_NONE, 0);
 }
 
 static void
@@ -2626,6 +2645,21 @@ _tf_stream_bus_message (TfStream *stream,
       cb_fs_component_state_changed (stream, component, fsstate);
       return TRUE;
     }
+  else if (gst_structure_has_name (s, "farsight-renegotiate"))
+    {
+      FsSession *fssession;
+      const GValue *value;
+
+      value = gst_structure_get_value (s, "session");
+      fssession = g_value_get_object (value);
+
+      if (fssession != stream->priv->fs_session)
+        return FALSE;
+
+      g_signal_emit (stream, signals[RESTART_SOURCE], 0);
+
+      return TRUE;
+    }
 
   return FALSE;
 }
@@ -2831,6 +2865,8 @@ _tf_stream_try_sending_codecs (TfStream *stream)
   GPtrArray *tpcodecs = NULL;
   GHashTable *feedback_messages = NULL;
   GPtrArray *header_extensions = NULL;
+  gboolean sent = FALSE;
+  GList *resend_codecs = NULL;
 
   DEBUG (stream, "called (send_local:%d send_supported:%d)",
       stream->priv->send_local_codecs, stream->priv->send_supported_codecs);
@@ -2875,6 +2911,7 @@ _tf_stream_try_sending_codecs (TfStream *stream)
           -1, tpcodecs, async_method_callback, "Media.StreamHandler::Ready",
           NULL, (GObject *) stream);
       stream->priv->send_local_codecs = FALSE;
+      sent = TRUE;
       goto out;
     }
 
@@ -2900,6 +2937,7 @@ _tf_stream_try_sending_codecs (TfStream *stream)
           -1, tpcodecs, async_method_callback,
           "Media.StreamHandler::SupportedCodecs", NULL, (GObject *) stream);
       stream->priv->send_supported_codecs = FALSE;
+      sent = TRUE;
 
       /* Fallthrough to potentially call CodecsUpdated as CMs assume
        * SupportedCodecs will only give the intersection of the already sent
@@ -2910,8 +2948,12 @@ _tf_stream_try_sending_codecs (TfStream *stream)
   /* Only send updates if there was something to update (iotw we sent codecs
    * before) or our list changed */
   if (stream->priv->last_sent_codecs != NULL
-      && !fs_codec_list_are_equal (fscodecs, stream->priv->last_sent_codecs))
+      && (resend_codecs =
+          fs_session_codecs_need_resend (stream->priv->fs_session,
+              stream->priv->last_sent_codecs, fscodecs)) != NULL)
     {
+      fs_codec_list_destroy (resend_codecs);
+
       if (!tpcodecs)
         tpcodecs = fs_codecs_to_tp (stream, fscodecs);
       if (!feedback_messages)
@@ -2935,6 +2977,7 @@ _tf_stream_try_sending_codecs (TfStream *stream)
           stream->priv->stream_handler_proxy,
           -1, tpcodecs, async_method_callback,
           "Media.StreamHandler::CodecsUpdated", NULL, (GObject *) stream);
+      sent = TRUE;
     }
 
 out:
@@ -2944,8 +2987,11 @@ out:
     g_boxed_free (TP_HASH_TYPE_RTCP_FEEDBACK_MESSAGE_MAP, feedback_messages);
   if (header_extensions)
     g_boxed_free (TP_ARRAY_TYPE_RTP_HEADER_EXTENSIONS_LIST, header_extensions);
-  fs_codec_list_destroy (stream->priv->last_sent_codecs);
-  stream->priv->last_sent_codecs = fscodecs;
+  if (sent)
+    {
+      fs_codec_list_destroy (stream->priv->last_sent_codecs);
+      stream->priv->last_sent_codecs = fscodecs;
+    }
 }
 
 /**
@@ -3014,6 +3060,7 @@ cb_fs_component_state_changed (TfStream *self,
     FsStreamState fsstate)
 {
   TpMediaStreamState state;
+  const gchar *state_str;
 
   if (component != 1)
     return;
@@ -3023,18 +3070,24 @@ cb_fs_component_state_changed (TfStream *self,
     case FS_STREAM_STATE_FAILED:
     case FS_STREAM_STATE_DISCONNECTED:
       state = TP_MEDIA_STREAM_STATE_DISCONNECTED;
+      state_str = "disconnected";
       break;
     case FS_STREAM_STATE_GATHERING:
     case FS_STREAM_STATE_CONNECTING:
     case FS_STREAM_STATE_CONNECTED:
+      state_str = "connecting";
       state = TP_MEDIA_STREAM_STATE_CONNECTING;
       break;
     case FS_STREAM_STATE_READY:
     default:
+      state_str = "connected";
       state = TP_MEDIA_STREAM_STATE_CONNECTED;
       break;
   }
 
+  DEBUG (self, "calling MediaStreamHandler::StreamState (%u: %s)", state,
+         state_str);
+
   self->priv->current_state = state;
 
   tp_cli_media_stream_handler_call_stream_state (
